import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np

π = torch.pi

### torch #################################

def nn_size(nn): return sum([p.numel() for p in nn.parameters()]) # get param count

def modulize(f):
	class Modulized(nn.Module):
		def __init__(self): super(Modulized,self).__init__()
		def forward(self, x): return f(x)
	return Modulized()

# https://towardsdatascience.com/extending-pytorch-with-custom-activation-functions-2d8b065ef2fa
def activation(f,grad_f=None):
	if grad_f==None:
		return modulize(f)
	class Activation(torch.autograd.Function):
		@staticmethod
		def forward(ctx, x):
			ctx.save_for_backward(x) # save input for backward pass
			return f(x)
		@staticmethod
		def backward(ctx, grad_output):
			grad_input = None # set output to None
			# if grad not required, return None to speed up computation
			if ctx.needs_input_grad[0]:
				# grad_input = (input >= 0).float() * grad_input
				x, = ctx.saved_tensors # restore input from context
				grad_input = grad_f(x,grad_output)
			return grad_input
	return modulize(Activation.apply)
	# return Activation.apply
# σ = activation(lambda x:x,lambda x,grad_out: grad_out)

### interpolations ########################

def smoothstep(t,v1,v2):
	λ = t * t * (3 - 2 * t)
	return v1*(1-λ) + v2*λ

def lerp(t,v1,v2):
	return v1*(1-t) + v2*t

### sampler ###############################

def sampler1d_smoothstep_repeat(tex,x):
	l = len(tex)
	X=x*l # so that x ∈ [0,1] wil cover the whole texture
	Xi = X.int() - (X<=0)*1
	t = X-Xi
	return smoothstep(t, tex[Xi % l], tex[(Xi + 1) % l])

def sampler1d_smoothstep_clamp(tex,x):
	l = len(tex)
	X=x*l # so that x ∈ [0,1] wil cover the whole texture
	Xi = X.int() - (X<=0)*1
	t = X-Xi
	return smoothstep(t, tex[torch.clamp(Xi,0,l-1)], tex[torch.clamp(Xi+1,0,l-1)])

def sampler1d_nearest_clamp(tex,x):
	l = len(tex)
	# if the coordinate are in [0, 1] replace them in the real image coordinate
	return tex[torch.clamp(torch.round(x*l-1).long(),0,l-1)]

def sampler1d_nearest_repeat(tex,x):
	l = len(tex)
	X=x*l # so that x ∈ [0,1] wil cover the whole texture
	Xi = (X+0.5).int() - (X<=0)*1
	return tex[Xi % l]

def sampler2d_nearest_clamp(tex, coord):
	c, h, w = tex.shape
	# if the coordinate are in [0, 1] replace them in the real image coordinate
	coord[:, 0] *= h-1
	coord[:, 1] *= w-1
	coord = torch.round(coord).long()
	coord_x = torch.clamp(coord[:, 1],0,w-1) # repeat last pixel
	coord_y = torch.clamp(coord[:, 0],0,h-1) # repeat last pixel
	pixels = tex[:, coord_y, coord_x]
	return pixels.permute(1,0)

### polynoms ##############################

# each components get a polynome applied
def poly(coeff):
	N = range(len(coeff))
	return lambda x: sum([coeff[n]*(x**n) for n in N])

# def poly(coeff):
# 	N = torch.arange(len(coeff))
# 	return lambda x: sum([coeff[n]*(x**n) for n in N])

# read channel[0] as x, else channel[i] as coefficients
# x assumed to be of shape [sample,channel,z_res,y_res,x_res]
#                       or [sample,channel,y_res,x_res]
#                       or [sample,channel,x_res]
#                       or [channel,x_res]
# def poly_chan(x): return None TODO

# def perlin1d_gen(res,seed=0):
# 	torch.manual_seed(seed)
# 	return 2*torch.rand(res).to(device)-1

### fake gradient | cantor ################

def cantor_function(x,n=6):
	if n==0: return x
	return (
		(((0<=x) & (x<=1/3))* cantor_function(3*x,n-1)/2 ) +
		((1/3<x) & (x<2/3)) * (1/2) +
		(2/3<=x)            * (1/2+cantor_function(3*x-2,n-1)/2)
	)

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
cantor_array = cantor_function(torch.linspace(0,1,3**12+1),12).to(device)

def cantor(): return modulize(lambda x: sampler1d_nearest_clamp(cantor_array,x))
# cantor = modulize(lambda x: torch.floor(x) + sampler1d_nearest_repeat(cantor_array,x))

def cantor2():
	return activation(
		lambda x: torch.floor(x) + sampler1d_nearest_repeat(cantor_array,x),
		lambda x,grad: grad                  # fake the gradient to identity
	)

### fake gradient | sigmoid staircase #####
### fake gradient | sigmoid sinus #########
### fake gradient | binary sinus ##########

### heterogenous ##########################

def multi(σs):
	def multi_(x):
		σcount = len(σs)
		out = torch.zeros_like(x)
		for (i,σ) in enumerate(σs):
			index = torch.arange(i,len(x)-1,σcount)
			out[index] = σ(x[index])
		return out
	return modulize(multi_)
